@@ -1,14 +1,7 @@
-{-# LANGUAGE BangPatterns        #-}
-{-# LANGUAGE LambdaCase          #-}
-{-# LANGUAGE ExplicitNamespaces  #-}
-{-# LANGUAGE NamedFieldPuns      #-}
-{-# LANGUAGE RankNTypes          #-}
-{-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE TupleSections       #-}
+{-# LANGUAGE ExplicitNamespaces    #-}
+{-# LANGUAGE QuantifiedConstraints #-}
 
 {-# OPTIONS_GHC -Wno-name-shadowing #-}
-{-# OPTIONS_GHC -Wno-unrecognised-pragmas #-}
-
 module Control.Monad.IOSim
   ( -- * Simulation monad
     IOSim
@@ -20,12 +13,14 @@ module Control.Monad.IOSim
   , Failure (..)
   , runSimTrace
   , runSimTraceST
+    -- *** QuickCheck Monadic combinators
+  , monadicIOSim_
+  , monadicIOSim
+  , runIOSimGen
     -- ** Explore races using /IOSimPOR/
     -- $iosimpor
   , exploreSimTrace
-  , exploreSimTraceST
   , controlSimTrace
-  , controlSimTraceST
   , ScheduleMod (..)
   , ScheduleControl (..)
     -- *** Exploration options
@@ -43,22 +38,17 @@ module Control.Monad.IOSim
   , unshareClock
     -- * Simulation trace
   , type SimTrace
-  , Trace (Cons,
-           Nil,
-           SimTrace,
-           SimPORTrace,
-           TraceDeadlock,
-           TraceLoop,
-           TraceMainReturn,
-           TraceMainException,
-           TraceRacesFound,
-           TraceInternalError)
+  , Time (..)
+  , Trace (Cons, Nil, SimTrace, SimPORTrace, TraceDeadlock, TraceLoop, TraceMainReturn, TraceMainException, TraceRacesFound, TraceInternalError)
+  , traceEvents
+  , traceResult
   , SimResult (..)
   , SimEvent (..)
   , SimEventType (..)
   , ThreadLabel
   , IOSimThreadId (..)
   , Labelled (..)
+  , Unique
     -- ** Dynamic Tracing
   , traceM
   , traceSTM
@@ -69,8 +59,6 @@ module Control.Monad.IOSim
   , ppSimEvent
   , ppDebug
     -- ** Selectors
-  , traceEvents
-  , traceResult
     -- *** list selectors
   , selectTraceEvents
   , selectTraceEvents'
@@ -100,39 +88,42 @@ module Control.Monad.IOSim
   , awaitTimeout
   ) where
 
-import           Prelude
+import Prelude
 
-import           Data.Bifoldable
-import           Data.Bifunctor (first)
-import           Data.Dynamic (fromDynamic)
-import           Data.Functor (void)
-import           Data.List (intercalate)
-import           Data.Maybe (catMaybes)
-import           Data.Set (Set)
-import qualified Data.Set as Set
-import           Data.Typeable (Typeable)
+import Data.Bifoldable
+import Data.Dynamic (fromDynamic)
+import Data.List (intercalate)
+import Data.Set (Set)
+import Data.Set qualified as Set
+import Data.Typeable (Typeable)
 
-import           Data.List.Trace (Trace (..))
-import qualified Data.List.Trace as Trace
+import Data.List.Trace (Trace (..))
 
-import           Control.Exception (throw)
+import Control.Exception (SomeAsyncException (..), throw)
 
-import           Control.Monad.ST.Lazy
-import           Data.STRef.Lazy
+import Control.Monad.ST.Lazy
 
-import           Control.Monad.Class.MonadThrow as MonadThrow
+import Control.Monad.Class.MonadThrow as MonadThrow
 
-import           Control.Monad.IOSim.Internal (runSimTraceST)
-import           Control.Monad.IOSim.Types
-import qualified Control.Monad.IOSimPOR.Internal as IOSimPOR (controlSimTraceST)
-import           Control.Monad.IOSimPOR.QuickCheckUtils
+import Control.Monad.IOSim.Internal (runSimTraceST)
+import Control.Monad.IOSim.Types
+import Control.Monad.IOSimPOR.Internal qualified as IOSimPOR (controlSimTraceST)
+import Control.Monad.IOSimPOR.QuickCheckUtils
 
-import           Test.QuickCheck
+import Test.QuickCheck
+import Test.QuickCheck.Gen.Unsafe (Capture (..), capture)
+import Test.QuickCheck.Monadic (PropertyM, monadic')
 
-import           System.IO.Unsafe
-import qualified Debug.Trace as Debug
+import System.IO.Unsafe
+
+import Data.Functor (void)
+import Data.IORef
+import Debug.Trace qualified as Debug
 
 
+-- | Select events according to the predicate function.  It throws an error if
+-- the simulation ends with 'Failure'.
+--
 selectTraceEvents
     :: (Time -> SimEventType -> Maybe b)
     -> SimTrace a
@@ -150,9 +141,13 @@ selectTraceEvents fn =
               []
     . traceSelectTraceEvents fn
 
+-- | Like 'selectTraceEvents', but it returns even if the simulation trace ends
+-- with 'Failure'.  It also works with any return type, not only `SimResult`
+-- like `selectTraceEvents` does.
+--
 selectTraceEvents'
     :: (Time ->  SimEventType -> Maybe b)
-    -> SimTrace a
+    -> Trace a SimEvent
     -> [b]
 selectTraceEvents' fn =
       bifoldr ( \ _ _   -> []  )
@@ -183,43 +178,34 @@ selectTraceRaces = go
 -- unsafe, of course, since that function may return different results
 -- at different times.
 
--- | Detach discovered races.  This is written in `ST` monad to support
--- simulations which do not terminate, in which case we will only detach races
--- up to the point we evaluated the simulation.
+detachTraceRaces :: forall a. Int -> SimTrace a -> (() -> [ScheduleControl], SimTrace a)
+detachTraceRaces debugLevel trace = unsafePerformIO $ do
+  races <- newIORef []
+  let readRaces :: () -> [ScheduleControl]
+      readRaces () = concat . reverse . unsafePerformIO $ readIORef races
+
+      saveRaces :: [ScheduleControl] -> x -> x
+      saveRaces rs t = unsafePerformIO $ modifyIORef races (rs:)
+                                      >> return t
+
+      go :: SimTrace a -> SimTrace a
+      go (SimTrace a b c d trace)      = traceDebugLog debugLevel (Left (SimEvent a b c d))
+                                       $ SimTrace a b c d $ go trace
+      go (SimPORTrace a b c d e trace) = traceDebugLog debugLevel (Left (SimPOREvent a b c d e))
+                                       $ SimPORTrace a b c d e $ go trace
+      go (TraceRacesFound rs trace)    = traceDebugLog debugLevel (Left (SimRacesFound rs))
+                                       $ saveRaces rs $ go trace
+      go (Cons a as)                   = traceDebugLog debugLevel (Left a)
+                                       $ Cons a (go as)
+      go (Nil a)                       = traceDebugLog debugLevel (Right a)
+                                       $ Nil a
+
+  return (readRaces, go trace)
+
+-- | Select all the traced values matching the expected type.  It relies on the
+-- sim's dynamic trace facility.
 --
-detachTraceRacesST :: forall a s. SimTrace a -> ST s (ST s [ScheduleControl], SimTrace a)
-detachTraceRacesST trace0 = do
-  races <- newSTRef []
-  let readRaces :: ST s [ScheduleControl]
-      readRaces = concat . reverse <$> readSTRef races
-
-      saveRaces :: [ScheduleControl] -> ST s ()
-      saveRaces rs = modifySTRef races (rs:)
-
-      go :: SimTrace a -> ST s (SimTrace a)
-      go (SimTrace a b c d trace)      = SimTrace a b c d <$> go trace
-      go (SimPORTrace _ _ _ _ EventRaces {} trace)
-                                       = go trace
-      go (SimPORTrace a b c d e trace) = SimPORTrace a b c d e <$> go trace
-      go (TraceRacesFound rs trace)    = saveRaces rs >> go trace
-      go t                             = return t
-
-  trace <- go trace0
-  return (readRaces, trace)
-
-
--- | Like `detachTraceRacesST`, but it doesn't expose discovered races.
---
-detachTraceRaces :: forall a. SimTrace a -> SimTrace a
-detachTraceRaces = Trace.filter (\a -> case a of
-                                         SimPOREvent { seType = EventRaces {} } -> False
-                                         SimRacesFound {} -> False
-                                         _                -> True)
-
--- | Select all the traced values matching the expected type. This relies on
--- the sim's dynamic trace facility.
---
--- For convenience, this throws exceptions for abnormal sim termination.
+-- For convenience, it throws exceptions for abnormal sim termination.
 --
 selectTraceEventsDynamic :: forall a b. Typeable b => SimTrace a -> [b]
 selectTraceEventsDynamic = selectTraceEvents fn
@@ -228,7 +214,8 @@ selectTraceEventsDynamic = selectTraceEvents fn
     fn _ (EventLog dyn) = fromDynamic dyn
     fn _ _              = Nothing
 
--- | Like 'selectTraceEventsDynamic' but also captures time of the trace event.
+-- | Like 'selectTraceEventsDynamic' but it also captures time of the trace
+-- event.
 --
 selectTraceEventsDynamicWithTime :: forall a b. Typeable b => SimTrace a -> [(Time, b)]
 selectTraceEventsDynamicWithTime = selectTraceEvents fn
@@ -237,19 +224,22 @@ selectTraceEventsDynamicWithTime = selectTraceEvents fn
     fn t (EventLog dyn) = (t,) <$> fromDynamic dyn
     fn _ _              = Nothing
 
--- | Like 'selectTraceEventsDynamic' but returns partial trace if an exception
--- is found in it.
+-- | Like 'selectTraceEventsDynamic' but it returns even if the simulation
+-- trace ends with 'Failure'.  It also works with any return type, not only
+-- `SimResult` like `selectTraceEventsDynamic` does.
 --
-selectTraceEventsDynamic' :: forall a b. Typeable b => SimTrace a -> [b]
+selectTraceEventsDynamic' :: forall a b. Typeable b => Trace a SimEvent -> [b]
 selectTraceEventsDynamic' = selectTraceEvents' fn
   where
     fn :: Time -> SimEventType -> Maybe b
     fn _ (EventLog dyn) = fromDynamic dyn
     fn _ _              = Nothing
 
--- | Like `selectTraceEventsDynamic'` but also captures time of the trace event.
+-- | Like `selectTraceEventsDynamicWithTime'` but it also captures time of the
+-- trace event.  It also works with any return type, not only `SimResult` like
+-- `selectTraceEventsDynamicWithTime` does.
 --
-selectTraceEventsDynamicWithTime' :: forall a b. Typeable b => SimTrace a -> [(Time, b)]
+selectTraceEventsDynamicWithTime' :: forall a b. Typeable b => Trace a SimEvent -> [(Time, b)]
 selectTraceEventsDynamicWithTime' = selectTraceEvents' fn
   where
     fn :: Time -> SimEventType -> Maybe (Time, b)
@@ -258,7 +248,7 @@ selectTraceEventsDynamicWithTime' = selectTraceEvents' fn
 
 -- | Get a trace of 'EventSay'.
 --
--- For convenience, this throws exceptions for abnormal sim termination.
+-- For convenience, it throws exceptions for abnormal sim termination.
 --
 selectTraceEventsSay :: SimTrace a -> [String]
 selectTraceEventsSay = selectTraceEvents fn
@@ -267,7 +257,7 @@ selectTraceEventsSay = selectTraceEvents fn
     fn _ (EventSay s) = Just s
     fn _ _            = Nothing
 
--- | Like 'selectTraceEventsSay' but also captures time of the trace event.
+-- | Like 'selectTraceEventsSay' but it also captures time of the trace event.
 --
 selectTraceEventsSayWithTime :: SimTrace a -> [(Time, String)]
 selectTraceEventsSayWithTime = selectTraceEvents fn
@@ -276,19 +266,20 @@ selectTraceEventsSayWithTime = selectTraceEvents fn
     fn t (EventSay s) = Just (t, s)
     fn _ _            = Nothing
 
--- | Like 'selectTraceEventsSay' but return partial trace if an exception is
--- found in it.
+-- | Like 'selectTraceEventsSay' but it returns even if the simulation trace
+-- ends with 'Failure'.  It also works with any return type, not only `SimResult`
+-- like `selectTraceEventsSay` does.
 --
-selectTraceEventsSay' :: SimTrace a -> [String]
+selectTraceEventsSay' :: Trace a SimEvent -> [String]
 selectTraceEventsSay' = selectTraceEvents' fn
   where
     fn :: Time -> SimEventType -> Maybe String
     fn _ (EventSay s) = Just s
     fn _ _            = Nothing
 
--- | Like `selectTraceEventsSay'` but also captures time of the trace event.
+-- | Like `selectTraceEventsSay'` but it also captures time of the trace event.
 --
-selectTraceEventsSayWithTime' :: SimTrace a -> [(Time, String)]
+selectTraceEventsSayWithTime' :: Trace a SimEvent -> [(Time, String)]
 selectTraceEventsSayWithTime' = selectTraceEvents' fn
   where
     fn :: Time -> SimEventType -> Maybe (Time, String)
@@ -297,18 +288,18 @@ selectTraceEventsSayWithTime' = selectTraceEvents' fn
 
 -- | Print all 'EventSay' to the console.
 --
--- For convenience, this throws exceptions for abnormal sim termination.
+-- For convenience, it throws exceptions for abnormal sim termination.
 --
 printTraceEventsSay :: SimTrace a -> IO ()
 printTraceEventsSay = mapM_ print . selectTraceEventsSay
 
 
--- | The most general select function.  It is a _total_ function.
+-- | The most general select function.  It is a /total function/.
 --
 traceSelectTraceEvents
     :: (Time -> SimEventType -> Maybe b)
-    -> SimTrace a
-    -> Trace (SimResult a) b
+    -> Trace a SimEvent
+    -> Trace a b
 traceSelectTraceEvents fn = bifoldr ( \ v _acc -> Nil v )
                                     ( \ eventCtx acc
                                      -> case eventCtx of
@@ -324,10 +315,11 @@ traceSelectTraceEvents fn = bifoldr ( \ v _acc -> Nil v )
                                     )
                                     undefined -- it is ignored
 
--- | Select dynamic events.  It is a _total_ function.
+-- | Select dynamic events.  It is a /total function/.
 --
 traceSelectTraceEventsDynamic :: forall a b. Typeable b
-                              => SimTrace a -> Trace (SimResult a) b
+                              => Trace a SimEvent
+                              -> Trace a b
 traceSelectTraceEventsDynamic = traceSelectTraceEvents fn
   where
     fn :: Time -> SimEventType -> Maybe b
@@ -335,9 +327,9 @@ traceSelectTraceEventsDynamic = traceSelectTraceEvents fn
     fn _ _              = Nothing
 
 
--- | Select say events.  It is a _total_ function.
+-- | Select say events.  It is a /total function/.
 --
-traceSelectTraceEventsSay :: forall a.  SimTrace a -> Trace (SimResult a) String
+traceSelectTraceEventsSay :: forall a. Trace a SimEvent -> Trace a String
 traceSelectTraceEventsSay = traceSelectTraceEvents fn
   where
     fn :: Time -> SimEventType -> Maybe String
@@ -394,7 +386,7 @@ instance Exception Failure where
              , "please report the issue at\n"
              , "https://github.com/input-output-hk/io-sim/issues"
              ]
-    
+
 
 -- | 'IOSim' is a pure monad.
 --
@@ -417,7 +409,7 @@ runSimOrThrow mainAction =
 runSimStrictShutdown :: forall a. (forall s. IOSim s a) -> Either Failure a
 runSimStrictShutdown mainAction = traceResult True (runSimTrace mainAction)
 
--- | Fold through the trace and return either a 'Failure' or the simulation
+-- | Fold through the trace and return either 'Failure' or a simulation
 -- result, i.e. the return value of the main thread.
 --
 traceResult :: Bool
@@ -430,7 +422,16 @@ traceResult strict = unsafePerformIO . eval
   where
     eval :: SimTrace a -> IO (Either Failure a)
     eval a = do
-      r <- try (evaluate a)
+      -- NOTE: It's fine to let asynchronous exceptions pass through. The only
+      -- way simulation could raise them is by using `throw` in pure code, while
+      -- `throwIO` in the simulation will be captured as `FailureException`.  So
+      -- we can safely assume asynchronous exceptions are coming from the
+      -- environment running the simulation, e.g. `QuickCheck`, as in the case
+      -- of `within` or `discardAfter` operators.
+      r <- tryJust (\e -> case fromException @SomeAsyncException e of
+                            Just _  -> Nothing
+                            Nothing -> Just e)
+                   (evaluate a)
       case r of
         Left e  -> return (Left (FailureEvaluation e))
         Right _ -> go a
@@ -502,19 +503,18 @@ runSimTrace mainAction = runST (runSimTraceST mainAction)
 -- slot.  In /IOSim/ and /IOSimPOR/ time only moves explicitly through timer
 -- events, e.g. things like `Control.Monad.Class.MonadTimer.SI.threadDelay`,
 -- `Control.Monad.Class.MonadTimer.SI.registerDelay` or the
--- `Control.Monad.Class.MonadTimer.NonStandard.MonadTimeout` API.  The usual
+-- `Control.Monad.Class.MonadTimer.MonadTimeout.NonStandard` API.  The usual
 -- QuickCheck techniques can help explore different schedules of
 -- threads too.
 
 -- | Execute a simulation, discover & revert races.  Note that this will execute
 -- the simulation multiple times with different schedules, and thus it's much
 -- more costly than a simple `runSimTrace` (also the simulation environments has
--- much more state to track and hence is slower).
+-- much more state to track and hence it is slower).
 --
 -- On property failure it will show the failing schedule (`ScheduleControl`)
--- which can be plugged to `controlSimTrace`.
---
--- Note: `exploreSimTrace` evaluates each schedule in parallel (using `par`).
+-- which can be passed to `controlSimTrace` to reproduce the failure without
+-- discovering the schedule.
 --
 exploreSimTrace
   :: forall a test. Testable test
@@ -526,78 +526,56 @@ exploreSimTrace
   -- ^ a callback which receives the previous trace (e.g. before reverting
   -- a race condition) and current trace
   -> Property
-exploreSimTrace optsf main k =
-    runST (exploreSimTraceST optsf main (\a b -> pure (k a b)))
-
-
--- | An 'ST' version of 'exploreSimTrace'. The callback also receives
--- 'ScheduleControl'.  This is mostly useful for testing /IOSimPOR/ itself.
---
--- Note: `exploreSimTraceST` evaluates each schedule sequentially.
---
-exploreSimTraceST
-  :: forall s a test. Testable test
-  => (ExplorationOptions -> ExplorationOptions)
-  -> (forall s. IOSim s a)
-  -> (Maybe (SimTrace a) -> SimTrace a -> ST s test)
-  -> ST s Property
-exploreSimTraceST optsf main k =
-    case explorationReplay opts of
-      Just control -> do
-        trace <- controlSimTraceST (explorationStepTimelimit opts) control main
-        counterexample ("Schedule control: " ++ show control)
-          <$> k Nothing trace
-      Nothing -> do
-        cacheRef <- createCacheST
-        prop <- go cacheRef (explorationScheduleBound opts)
-                            (explorationBranching opts)
-                            ControlDefault Nothing
-        !size <- cacheSizeST cacheRef
-        return $ tabulate "Modified schedules explored" [bucket size] prop
+exploreSimTrace optsf mainAction k =
+  case explorationReplay opts of
+    Nothing ->
+      explore (explorationScheduleBound opts) (explorationBranching opts) ControlDefault Nothing .&&.
+      let size = cacheSize() in size `seq`
+      tabulate "Modified schedules explored" [bucket size] True
+    Just control ->
+      replaySimTrace opts mainAction control (k Nothing)
   where
     opts = optsf stdExplorationOptions
 
-    go :: STRef s (Set ScheduleControl)
-       -> Int -- schedule bound
-       -> Int -- branching factor
-       -> ScheduleControl
-       -> Maybe (SimTrace a)
-       -> ST s Property
-    go cacheRef n m control passingTrace = do
-      traceWithRaces <- IOSimPOR.controlSimTraceST (explorationStepTimelimit opts) control main
-      (readRaces, trace0) <- detachTraceRacesST traceWithRaces
-      (readSleeperST, trace) <- compareTracesST passingTrace trace0
-      () <- traceDebugLog (explorationDebugLevel opts) traceWithRaces
-      conjoinNoCatchST
-        [ do sleeper <- readSleeperST
-             prop <- k passingTrace trace
-             return $ counterexample ("Schedule control: " ++ show control)
-                    $ counterexample
-                       (case sleeper of
-                         Nothing -> "No thread delayed"
-                         Just ((t,tid,lab),racing) ->
-                           showThread (tid,lab) ++
-                           " delayed at time "++
-                           show t ++
-                           "\n  until after:\n" ++
-                           unlines (map (("    "++).showThread) $ Set.toList racing)
-                        )
-                      prop
-        , do let limit = (n+m-1) `div` m
-              -- To ensure the set of schedules explored is deterministic, we
-              -- filter out cached ones *after* selecting the children of this
-              -- node.
-             races <- catMaybes
-                  <$> (readRaces >>= traverse (cachedST cacheRef) . take limit)
-             let branching = length races
-             -- tabulate "Races explored" (map show races) $
-             tabulate "Branching factor" [bucket branching]
-                .  tabulate "Race reversals per schedule" [bucket (raceReversals control)]
-                .  conjoin
-               <$> sequence
-                     [ go cacheRef n' ((m-1) `max` 1) r (Just trace0)
-                     | (r,n') <- zip races (divide (n-branching) branching) ]
-        ]
+    explore :: Int -- schedule bound
+            -> Int -- branching factor
+            -> ScheduleControl -> Maybe (SimTrace a) -> Property
+    explore n m control passingTrace =
+
+      -- ALERT!!! Impure code: readRaces must be called *after* we have
+      -- finished with trace.
+      let (readRaces, trace0) = detachTraceRaces (explorationDebugLevel opts)
+                              $ controlSimTrace
+                                  (explorationStepTimelimit opts) control mainAction
+          (sleeper,trace) = compareTraces passingTrace trace0
+      in ( counterexample ("Schedule control: " ++ show control)
+         $ counterexample
+            (case sleeper of
+              Nothing -> "No thread delayed"
+              Just ((t,tid,lab),racing) ->
+                showThread (tid,lab) ++
+                " delayed at time "++
+                show t ++
+                "\n  until after:\n" ++
+                unlines (map (("    "++).showThread) $ Set.toList racing)
+             )
+         $ k passingTrace trace
+         )
+      .&&| let limit     = (n+m-1) `div` m
+               -- To ensure the set of schedules explored is deterministic, we
+               -- filter out cached ones *after* selecting the children of this
+               -- node.
+               races     = filter (not . cached) . take limit $ readRaces ()
+               branching = length races
+           in -- tabulate "Races explored" (map show races) $
+              tabulate "Branching factor" [bucket branching] $
+              tabulate "Race reversals per schedule" [bucket (raceReversals control)] $
+              conjoinPar
+                [ --Debug.trace "New schedule:" $
+                  --Debug.trace ("  "++show r) $
+                  --counterexample ("Schedule control: " ++ show r) $
+                  explore n' ((m-1) `max` 1) r (Just trace0)
+                | (r,n') <- zip races (divide (n-branching) branching) ]
 
     bucket :: Int -> String
     bucket n | n<10      = show n
@@ -616,42 +594,69 @@ exploreSimTraceST optsf main k =
       ppIOSimThreadId tid ++ (case lab of Nothing -> ""
                                           Just l  -> " ("++l++")")
 
+    -- cache of explored schedules
+    cache :: IORef (Set ScheduleControl)
+    cache = unsafePerformIO cacheIO
+
     -- insert a schedule into the cache
-    cachedST :: STRef s (Set ScheduleControl) -> ScheduleControl -> ST s (Maybe ScheduleControl)
-    cachedST cacheRef a = do
-      set <- readSTRef cacheRef
-      writeSTRef cacheRef (Set.insert a set)
-      return $ if Set.member a set
-               then Nothing
-               else Just a
+    cached :: ScheduleControl -> Bool
+    cached = unsafePerformIO . cachedIO
+
+    -- compute cache size; it's a function to make sure that `GHC` does not
+    -- inline it (and share the same thunk).
+    cacheSize :: () -> Int
+    cacheSize = unsafePerformIO . cacheSizeIO
 
     --
-    -- Caching in ST monad
+    -- Caching in IO monad
     --
 
     -- It is possible for the same control to be generated several times.
     -- To avoid exploring them twice, we keep a cache of explored schedules.
-    createCacheST :: ST s (STRef s (Set ScheduleControl))
-    createCacheST = newSTRef $
+    cacheIO :: IO (IORef (Set ScheduleControl))
+    cacheIO = newIORef $
               -- we use opts here just to be sure the reference cannot be
               -- lifted out of exploreSimTrace
               if explorationScheduleBound opts >=0
                 then Set.empty
                 else error "exploreSimTrace: negative schedule bound"
 
-    cacheSizeST :: STRef s (Set ScheduleControl) -> ST s Int
-    cacheSizeST = fmap Set.size . readSTRef
+    cachedIO :: ScheduleControl -> IO Bool
+    cachedIO m = atomicModifyIORef' cache $ \set ->
+      (Set.insert m set, Set.member m set)
+
+    cacheSizeIO :: () -> IO Int
+    cacheSizeIO () = Set.size <$> readIORef cache
 
 
 -- |  Trace `SimTrace` to `stderr`.
 --
-traceDebugLog :: Int -> SimTrace a -> ST s ()
-traceDebugLog logLevel _trace | logLevel <= 0 = pure ()
-traceDebugLog 1 trace = Debug.traceM $ "Simulation trace with discovered schedules:\n"
-                                    ++ Trace.ppTrace (ppSimResult 0 0 0) (ppSimEvent 0 0 0) (ignoreRaces $ void `first` trace)
-traceDebugLog _ trace = Debug.traceM $ "Simulation trace with discovered schedules:\n"
-                                    ++ Trace.ppTrace (ppSimResult 0 0 0) (ppSimEvent 0 0 0) (void `first` trace)
+-- An internal function.
+--
+traceDebugLog :: Int -> Either SimEvent (SimResult a) -> SimTrace a -> SimTrace a
+traceDebugLog logLevel _event trace | logLevel <= 0 = trace
+-- Discard races if log level is 1
+traceDebugLog 1 (Left SimPOREvent { seType = EventRaces {} }) trace = trace
+traceDebugLog 1 (Left event) trace  = Debug.trace (ppSimEvent 0 0 0 event) trace
+traceDebugLog 1 (Right event) trace = Debug.trace (ppSimResult 0 0 0 (void event)) trace
+-- Otherwise, show races
+traceDebugLog _ (Left event) trace  = Debug.trace (ppSimEvent 0 0 0 event) trace
+traceDebugLog _ (Right event) trace = Debug.trace (ppSimResult 0 0 0 (void event)) trace
 
+replaySimTrace :: forall a test. (Testable test)
+               => ExplorationOptions
+               -- ^ race exploration options
+               -> (forall s. IOSim s a)
+               -> ScheduleControl
+               -- ^ a schedule control to reproduce
+               -> (SimTrace a -> test)
+               -- ^ a callback which receives the simulation trace. The trace
+               -- will not contain any race events
+               -> Property
+replaySimTrace opts mainAction control k =
+  let (_,trace) = detachTraceRaces (explorationDebugLevel opts) $
+                  controlSimTrace (explorationStepTimelimit opts) control mainAction
+  in property (k trace)
 
 -- | Run a simulation using a given schedule.  This is useful to reproduce
 -- failing cases without exploring the races.
@@ -668,28 +673,18 @@ controlSimTrace :: forall a.
                 -- ^ a simulation to run
                 -> SimTrace a
 controlSimTrace limit control main =
-    runST (controlSimTraceST limit control main)
-
-controlSimTraceST :: Maybe Int -> ScheduleControl -> IOSim s a -> ST s (SimTrace a)
-controlSimTraceST limit control main =
-  detachTraceRaces <$> IOSimPOR.controlSimTraceST limit control main
-
+    runST (IOSimPOR.controlSimTraceST limit control main)
 
 --
 -- Utils
 --
 
-ignoreRaces :: SimTrace a -> SimTrace a
-ignoreRaces = Trace.filter (\a -> case a of
-                                    SimPOREvent { seType = EventRaces {} } -> False
-                                    _ -> True)
-
 raceReversals :: ScheduleControl -> Int
 raceReversals ControlDefault      = 0
 raceReversals (ControlAwait mods) = length mods
 raceReversals ControlFollow{}     = error "Impossible: raceReversals ControlFollow{}"
 
--- compareTraces is given (maybe) a passing trace and a failing trace,
+-- `compareTracesST` is given (maybe) a passing trace and a failing trace,
 -- and identifies the point at which they diverge, where it inserts a
 -- "sleep" event for the thread that is delayed in the failing case,
 -- and a "wake" event before its next action. It also returns the
@@ -697,61 +692,112 @@ raceReversals ControlFollow{}     = error "Impossible: raceReversals ControlFoll
 -- to be consumed lazily (and perhaps only partially), and since the
 -- sleeping thread is not of interest unless the trace is consumed
 -- this far, then we collect its identity only if it is reached using
--- unsafePerformIO.
+-- `unsafePerformIO`.
 
--- TODO: return StepId
-compareTracesST :: forall a b s.
-                   Maybe (SimTrace a) -- ^ passing
-                -> SimTrace b         -- ^ failing
-                -> ST s ( ST s (Maybe ( (Time, IOSimThreadId, Maybe ThreadLabel)
-                                      , Set.Set (IOSimThreadId, Maybe ThreadLabel)
-                                      ))
-                        , SimTrace b
-                        )
-compareTracesST Nothing trace = return (return Nothing, trace)
-compareTracesST (Just passing) trace = do
-  sleeper <- newSTRef Nothing
-  trace' <- go sleeper passing trace
-  return ( readSTRef sleeper
-         , trace'
-         )
-  where
-        go :: STRef s (Maybe ( (Time, IOSimThreadId, Maybe ThreadLabel)
-                             , Set.Set (IOSimThreadId, Maybe ThreadLabel)
-                             ))
-           -> SimTrace a -- ^ passing
-           -> SimTrace b -- ^ failing
-           -> ST s (SimTrace b)
-        go sleeper (SimPORTrace tpass tidpass _ _ _ pass')
+compareTraces :: Maybe (SimTrace a1)
+              -> SimTrace a2
+              -> (Maybe ((Time, IOSimThreadId, Maybe ThreadLabel),
+                         Set.Set (IOSimThreadId, Maybe ThreadLabel)),
+                  SimTrace a2)
+compareTraces Nothing trace = (Nothing, trace)
+compareTraces (Just passing) trace = unsafePerformIO $ do
+  sleeper <- newIORef Nothing
+  return (unsafePerformIO $ readIORef sleeper,
+          go sleeper passing trace)
+  where go sleeper (SimPORTrace tpass tidpass _ _ _ pass')
                    (SimPORTrace tfail tidfail tstepfail tlfail evfail fail')
           | (tpass,tidpass) == (tfail,tidfail) =
               SimPORTrace tfail tidfail tstepfail tlfail evfail
-                <$> go sleeper pass' fail'
-        go sleeper (SimPORTrace tpass tidpass tsteppass tlpass _ _) fail = do
-          writeSTRef sleeper $ Just ((tpass, tidpass, tlpass),Set.empty)
-          SimPORTrace tpass tidpass tsteppass tlpass EventThreadSleep
-            <$> wakeup sleeper tidpass fail
-        go _ SimTrace {} _ = error "compareTracesST: invariant violation"
-        go _ _ SimTrace {} = error "compareTracesST: invariant violation"
-        go _ _ fail = return fail
+                $ go sleeper pass' fail'
+        go sleeper (SimPORTrace tpass tidpass tsteppass tlpass _ _) fail =
+          unsafePerformIO $ do
+            writeIORef sleeper $ Just ((tpass, tidpass, tlpass),Set.empty)
+            return $ SimPORTrace tpass tidpass tsteppass tlpass EventThreadSleep
+                   $ wakeup sleeper tidpass fail
+        go _ SimTrace {} _ = error "compareTraces: invariant violation"
+        go _ _ SimTrace {} = error "compareTraces: invariant violation"
+        go _ _ fail = fail
 
-        wakeup :: STRef s (Maybe ( (Time, IOSimThreadId, Maybe ThreadLabel)
-                                 , Set.Set (IOSimThreadId, Maybe ThreadLabel)
-                                 ))
-               -> IOSimThreadId
-               -> SimTrace b
-               -> ST s (SimTrace b)
         wakeup sleeper tidpass
                fail@(SimPORTrace tfail tidfail tstepfail tlfail evfail fail')
           | tidpass == tidfail =
-              return $ SimPORTrace tfail tidfail tstepfail tlfail EventThreadWake fail
-          | otherwise = do
-              ms <- readSTRef sleeper
-              case ms of
-                Just (slp, racing) -> do
-                  writeSTRef sleeper $ Just (slp,Set.insert (tidfail,tlfail) racing)
-                  SimPORTrace tfail tidfail tstepfail tlfail evfail
-                    <$> wakeup sleeper tidpass fail'
-                Nothing -> error "compareTraceST: invariant violation"
-        wakeup _ _ SimTrace {} = error "compareTracesST: invariant violation"
-        wakeup _ _ fail = return fail
+              SimPORTrace tfail tidfail tstepfail tlfail EventThreadWake fail
+          | otherwise = unsafePerformIO $ do
+              Just (slp,racing) <- readIORef sleeper
+              writeIORef sleeper $ Just (slp,Set.insert (tidfail,tlfail) racing)
+              return $ SimPORTrace tfail tidfail tstepfail tlfail evfail
+                     $ wakeup sleeper tidpass fail'
+        wakeup _ _ SimTrace {} = error "compareTraces: invariant violation"
+        wakeup _ _ fail = fail
+
+--
+-- QuickCheck monadic combinators
+--
+
+
+
+-- | Like <monadicST https://hackage.haskell.org/package/QuickCheck-2.14.3/docs/Test-QuickCheck-Monadic.html#v:monadicST>.
+--
+-- Note: it calls `traceResult` in non-strict mode, e.g. leaked threads do not
+-- cause failures.
+--
+-- @since 1.4.1.0
+--
+monadicIOSim_ :: Testable a
+              => (forall s. PropertyM (IOSim s) a)
+              -> Property
+monadicIOSim_ sim =
+    monadicIOSim
+      (\e -> case traceResult False e of
+          Left e  -> counterexample (show e) False
+          Right p -> p)
+      id
+      sim
+
+-- | A more general version of `monadicIOSim_`, which:
+--
+-- * allows to run in monad stacks build on top of `IOSim`;
+-- * gives more control how to attach debugging information to failed
+--   tests.
+--
+-- Note, to use this combinator your monad needs to be defined as:
+--
+-- > newtype M s a = M s { runM :: ReaderT State (IOSim s) a }
+--
+-- It's important that `M s` is a monad.  For such a monad one you'll need provide
+-- a natural transformation:
+-- @
+--   -- the state could also be created as an `IOSim` computation.
+--   nat :: forall s a. State -> M s a -> 'IOSim' s a
+--   nat state m = runStateT (runM m) state
+-- @
+--
+-- @since 1.4.1.0
+--
+monadicIOSim :: (Testable a, forall s. Monad (m s))
+             => (SimTrace Property -> Property)
+             -- ^ Allows to trace `SimTrace` in counterexamples.  The simplest
+             -- use case is to pass:
+             --
+             -- > either (\e -> counterexample (show e) False) id . traceResult False
+             --
+             -- as `monadicIOSim_` does.
+             --
+             -> (forall s a. m s a -> IOSim s a)
+             -- ^ natural transformation from `m` to @IOSim` s@
+             -> (forall s. PropertyM (m s) a)
+             -> Property
+monadicIOSim f tr sim = property (runIOSimGen f (tr <$> monadic' sim))
+
+-- | Like <runSTGen
+-- https://hackage.haskell.org/package/QuickCheck-2.14.3/docs/Test-QuickCheck-Monadic.html#v:runSTGen>.
+--
+-- @since 1.4.1.0
+--
+runIOSimGen :: (SimTrace a -> Property)
+            -> (forall s. Gen (IOSim s a))
+            -> Gen Property
+runIOSimGen f sim = do
+    Capture eval <- capture
+    let trace = runSimTrace (eval sim)
+    return (f trace)
