@@ -1,12 +1,15 @@
-{-# LANGUAGE LambdaCase #-}
-
+{-# LANGUAGE CPP #-}
 -- | A minimal implementation of a strict deque.
 --
 module Data.Deque.Strict where
 
-import           Prelude hiding (head, init, tail)
-import           Data.Foldable (foldl', foldr')
-import qualified Data.List as List
+#if __GLASGOW_HASKELL__ >= 910
+import Data.Foldable (foldr')
+#else
+import Data.Foldable (foldl', foldr')
+#endif
+import Data.List qualified as List
+import Prelude hiding (head, init, tail)
 
 data Deque a = Deque ![a] ![a]
 
