@@ -1,12 +1,9 @@
-{-# LANGUAGE CPP                 #-}
-{-# LANGUAGE ConstraintKinds     #-}
-{-# LANGUAGE DeriveFunctor       #-}
-{-# LANGUAGE FlexibleContexts    #-}
-{-# LANGUAGE NumericUnderscores  #-}
-{-# LANGUAGE RankNTypes          #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE CPP #-}
 
 {-# OPTIONS_GHC -Wno-orphans #-}
+-- `-fno-full-laziness` is needed for `discardAfter` to work correctly, see
+-- `unit_discardAfter` below.
+{-# OPTIONS_GHC -fno-full-laziness #-}
 
 module Test.Control.Monad.IOSim
   ( tests
@@ -22,33 +19,36 @@ module Test.Control.Monad.IOSim
   , singleTimeoutExperiment
   ) where
 
-import           Data.Either (isLeft)
-import           Data.Fixed (Micro)
-import           Data.Foldable (foldl')
-import           Data.Functor (($>))
-import           Data.Time.Clock (picosecondsToDiffTime)
+import Data.Either (isLeft)
+import Data.Fixed (Micro)
+#if __GLASGOW_HASKELL__ < 910
+import Data.Foldable (foldl')
+#endif
+import Data.Functor (($>))
+import Data.Time.Clock (picosecondsToDiffTime)
 
-import           Control.Exception (ArithException (..), AsyncException)
-import           Control.Monad
-import           Control.Monad.Fix
-import           System.IO.Error (ioeGetErrorString, isUserError)
+import Control.Exception (ArithException (..), AsyncException)
+import Control.Monad
+import Control.Monad.Fix
+import System.IO.Error (ioeGetErrorString, isUserError)
 
-import           Control.Concurrent.Class.MonadSTM.Strict
-import qualified Control.Concurrent.Class.MonadSTM.TVar as LazySTM
-import           Control.Monad.Class.MonadAsync
-import           Control.Monad.Class.MonadFork
-import           Control.Monad.Class.MonadSay
-import           Control.Monad.Class.MonadThrow
-import           Control.Monad.Class.MonadTime.SI
-import           Control.Monad.Class.MonadTimer.SI
-import           Control.Monad.IOSim
+import Control.Concurrent.Class.MonadSTM.Strict
+import Control.Concurrent.Class.MonadSTM.TVar qualified as LazySTM
+import Control.Monad.Class.MonadAsync
+import Control.Monad.Class.MonadFork
+import Control.Monad.Class.MonadSay
+import Control.Monad.Class.MonadThrow
+import Control.Monad.Class.MonadTime.SI
+import Control.Monad.Class.MonadTimer.SI
+import Control.Monad.IOSim
 
-import           Test.Control.Monad.Utils
-import           Test.Control.Monad.STM
+import Test.Control.Monad.STM
+import Test.Control.Monad.Utils
 
-import           Test.QuickCheck
-import           Test.Tasty hiding (after)
-import           Test.Tasty.QuickCheck
+import Test.QuickCheck
+import Test.QuickCheck.Property as QC
+import Test.Tasty hiding (after)
+import Test.Tasty.QuickCheck
 
 tests :: TestTree
 tests =
@@ -68,8 +68,13 @@ tests =
     , testProperty "async exceptions 2"     unit_timeouts_and_async_exceptions_2
     , testProperty "async exceptions 3"     unit_timeouts_and_async_exceptions_3
     , testProperty "threadDelay and STM"    unit_threadDelay_and_stm
+    , testProperty "{register,thread}Delay" unit_registerDelay_threadDelay
     , testProperty "throwTo and STM"        unit_throwTo_and_stm
     ]
+  , testGroup "QuickCheck"
+    [ testProperty "timeout: discardAfter"  unit_discardAfter
+    , testProperty "timeout: within"        unit_within
+    ]
   , testProperty "threadId order (IOSim)"   (withMaxSuccess 1000 prop_threadId_order_order_Sim)
   , testProperty "forkIO order (IOSim)"     (withMaxSuccess 1000 prop_fork_order_ST)
   , testProperty "order (IO)"               (expectFailure prop_fork_order_IO)
@@ -176,6 +181,24 @@ tests =
     , testProperty "registerDelayCancellable (IO impl)"
         prop_registerDelayCancellable_IO
     ]
+  , testGroup "MonadSTM"
+    [ testGroup "flushTQueue"
+      [ testProperty "empties the queue IO" prop_flushTQueueEmpties_IO
+      , testProperty "empties the queue IOSim" prop_flushTQueueEmpties_IOSim
+      , testProperty "maintains FIFO order IO" prop_flushTQueueOrder_IO
+      , testProperty "maintains FIFO order IOSim" prop_flushTQueueOrder_IOSim
+      ]
+    , testGroup "flushTBQueue"
+      [ testProperty "empties the queue IO" prop_flushTBQueueEmpties_IO
+      , testProperty "empties the queue IOSim" prop_flushTBQueueEmpties_IOSim
+      , testProperty "maintains FIFO order IO" prop_flushTBQueueOrder_IO
+      , testProperty "maintains FIFO order IOSim" prop_flushTBQueueOrder_IOSim
+      ]
+    , testGroup "tryReadTBQueue"
+      [ testProperty "works correctly when the queue is empty IO" prop_tryReadEmptyTBQueue_IO
+      , testProperty "works correctly when the queue is empty IOSim" prop_tryReadEmptyTBQueue_IOSim
+      ]
+    ]
   ]
 
 --
@@ -872,7 +895,6 @@ type TimeoutConstraints m =
       , MonadMask  m
       , MonadThrow (STM m)
       , MonadSay   m
-      , MonadMaskingState m
       )
 
 instance Arbitrary DiffTime where
@@ -995,7 +1017,7 @@ prop_timeout
     :: TimeoutDuration
     -> ActionDuration
     -> Property
-prop_timeout intendedTimeoutDuration intendedActionDuration = 
+prop_timeout intendedTimeoutDuration intendedActionDuration =
     runSimOrThrow (withSanityCheck <$>
                     singleTimeoutExperiment
                       intendedTimeoutDuration
@@ -1059,6 +1081,46 @@ prop_stacked_timeouts timeout0 timeout1 actionDuration =
               = Just Nothing
 
 
+-- | Check that `discardAfter` works as expected.
+--
+-- NOTE: using `discardAfter` with `IOSim` is more tricky than for `IO`
+-- properties, since `IOSim` is a pure computation.  One need to wrap the
+-- simulation in a lambda and use `-fno-full-laziness` to avoid GHC from
+-- moving the thunk outside of the lambda, and evaluating it just once.
+--
+unit_discardAfter :: Property
+unit_discardAfter = mapTotalResult f
+                  . discardAfter 10
+                  $ \() -> runSimOrThrow $ True <$ (forever (threadDelay 10))
+  where
+    -- if `discard` kills the computation with the `Timeout` exception,
+    -- `theException` is `Nothing`, but if `traceResult` wraps it, then it is
+    -- a `Just`. We mark each test a success if `theException` is `Nothing`,
+    -- otherwise the test would fail with too many discarded cases, but if we re
+    -- introduce the bug in `traceResult` then it fails, since then
+    -- `theException` is a `Just`.
+    f :: QC.Result -> QC.Result
+    f r@MkResult { QC.theException = Nothing }
+        = r { ok = Just True }
+    f r = r
+
+
+-- | Check that `within` works as expected.
+--
+unit_within :: Property
+unit_within = mapTotalResult f
+            . within 10
+            $ runSimOrThrow $ True <$ (forever (threadDelay 10))
+  where
+    -- if `within` kills the computation with the `Timeout` exception,
+    -- `theException` is `Nothing`, but if `traceResult` wraps it, then it is
+    -- a `Just`.
+    f :: QC.Result -> QC.Result
+    f r@MkResult { QC.theException = Nothing }
+        = r { expect = False }
+    f r = r
+
+
 unit_timeouts_and_async_exceptions_1 :: Property
 unit_timeouts_and_async_exceptions_1 =
     let trace = runSimTrace experiment in
@@ -1077,7 +1139,7 @@ unit_timeouts_and_async_exceptions_1 =
       threadDelay (delay / 2)
       killThread tid
       threadDelay 1
-      return $ property True 
+      return $ property True
 
 
 unit_timeouts_and_async_exceptions_2 :: Property
@@ -1098,7 +1160,7 @@ unit_timeouts_and_async_exceptions_2 =
       threadDelay (delay / 2)
       killThread tid
       threadDelay 1
-      return $ property True 
+      return $ property True
 
 
 unit_timeouts_and_async_exceptions_3 :: Property
@@ -1119,7 +1181,7 @@ unit_timeouts_and_async_exceptions_3 =
       threadDelay (delay / 2)
       killThread tid
       threadDelay 1
-      return $ property True 
+      return $ property True
 
 
 -- | Verify that a thread blocked on `threadDelay` is not unblocked by an STM
@@ -1153,6 +1215,34 @@ unit_threadDelay_and_stm =
 
       return (t1 `diffTime` t0 === delay)
 
+unit_registerDelay_threadDelay :: Property
+unit_registerDelay_threadDelay =
+    let trace = runSimTrace experiment in
+        counterexample (ppTrace_ trace)
+      . either (\e -> counterexample (show e) False) id
+      . traceResult False
+      $ trace
+  where
+    experiment :: IOSim s Property
+    experiment = do
+      v0 <- registerDelay 2
+      v1 <- newTVarIO False
+
+      _ <- forkIO $ do
+        threadDelay 1
+        atomically $ writeTVar v1 True
+
+      atomically $ do
+        b0 <- LazySTM.readTVar v0
+        b1 <- readTVar v1
+        check $ b0 || b1
+
+      let delay = 2
+      t0 <- getMonotonicTime
+      threadDelay delay
+      t1 <- getMonotonicTime
+
+      return (t1 `diffTime` t0 === delay)
 
 -- | Verify that a thread blocked on `throwTo` is not unblocked by an STM
 -- transaction.
@@ -1289,7 +1379,7 @@ instance Arbitrary DelayWithCancel where
 
 prop_registerDelayCancellable
     :: (forall s. DiffTime -> IOSim s (STM (IOSim s) TimeoutState, IOSim s ()))
-    -- ^ implementation 
+    -- ^ implementation
     -> DelayWithCancel
     -> Property
 prop_registerDelayCancellable registerDelayCancellableImpl
@@ -1348,6 +1438,95 @@ prop_registerDelayCancellable_IO =
         cancelTimeout
         awaitTimeout
 
+-- | Test that 'flushTQueue' empties the queue in `IO`.
+prop_flushTQueueEmpties_IO :: Property
+prop_flushTQueueEmpties_IO =
+  ioProperty emptyTQueueAfterFlush
+
+-- | Test that 'flushTQueue' empties the queue in `IOSim`.
+prop_flushTQueueEmpties_IOSim :: Bool
+prop_flushTQueueEmpties_IOSim =
+  runSimOrThrow emptyTQueueAfterFlush
+
+emptyTQueueAfterFlush :: MonadSTM m => m Bool
+emptyTQueueAfterFlush = do
+  q <- newTQueueIO
+  atomically $ do
+    writeTQueue q (1 :: Int)
+    _ <- flushTQueue q
+    isEmptyTQueue q
+
+-- | Test that 'flushTQueue' returns values in FIFO order in `IO`.
+prop_flushTQueueOrder_IO :: [Int] -> Property
+prop_flushTQueueOrder_IO entries =
+  ioProperty (writeAndFlushTQueue entries >>= \actual -> pure $ actual === entries)
+
+-- | Test that 'flushTQueue' returns values in FIFO order in `IOSim`.
+prop_flushTQueueOrder_IOSim :: [Int] -> Property
+prop_flushTQueueOrder_IOSim entries =
+  runSimOrThrow (writeAndFlushTQueue entries) === entries
+
+writeAndFlushTQueue :: MonadSTM m => [Int] -> m [Int]
+writeAndFlushTQueue entries =
+  atomically $ do
+    q <- newTQueue
+    forM_ entries $ writeTQueue q
+    flushTQueue q
+
+-- | Test that 'flushTBQueue' empties the queue in `IO`.
+prop_flushTBQueueEmpties_IO :: Property
+prop_flushTBQueueEmpties_IO =
+  ioProperty emptyTBQueueAfterFlush
+
+-- | Test that 'flushTBQueue' empties the queue in `IOSim`.
+prop_flushTBQueueEmpties_IOSim :: Bool
+prop_flushTBQueueEmpties_IOSim =
+  runSimOrThrow emptyTBQueueAfterFlush
+
+emptyTBQueueAfterFlush :: MonadSTM m => m Bool
+emptyTBQueueAfterFlush = do
+  q <- newTBQueueIO 10
+  atomically $ do
+    writeTBQueue q (1 :: Int)
+    writeTBQueue q 2
+    _ <- flushTBQueue q
+    isEmptyTBQueue q
+
+-- | Test that 'flushTBQueue' returns values in FIFO order in `IO`.
+prop_flushTBQueueOrder_IO :: [Int] -> Property
+prop_flushTBQueueOrder_IO entries =
+  ioProperty (writeAndFlushTBQueue entries >>= \actual -> pure $ actual === entries)
+
+-- | Test that 'flushTBQueue' returns values in FIFO order in `IOSim`.
+prop_flushTBQueueOrder_IOSim :: [Int] -> Property
+prop_flushTBQueueOrder_IOSim entries =
+  runSimOrThrow (writeAndFlushTBQueue entries) === entries
+
+writeAndFlushTBQueue :: MonadSTM m => [Int] -> m [Int]
+writeAndFlushTBQueue entries =
+  atomically $ do
+    q <- newTBQueue (1 + fromIntegral (length entries))
+    forM_ entries $ writeTBQueue q
+    flushTBQueue q
+
+prop_tryReadEmptyTBQueue_IO :: Bool -> Property
+prop_tryReadEmptyTBQueue_IO sndRead =
+  ioProperty $ tryReadEmptyTBQueue sndRead
+
+prop_tryReadEmptyTBQueue_IOSim :: Bool -> Property
+prop_tryReadEmptyTBQueue_IOSim sndRead =
+  runSimOrThrow $ tryReadEmptyTBQueue sndRead
+
+tryReadEmptyTBQueue :: MonadSTM m => Bool -> m Property
+tryReadEmptyTBQueue sndRead = atomically $ do
+  q <- newTBQueue 10
+  _ <- tryReadTBQueue q
+  writeTBQueue q ()
+  when sndRead $ void $ tryReadTBQueue q
+  l <- lengthTBQueue q
+
+  pure $ l === if sndRead then 0 else 1
+
 --
 -- Utils
 --
