@@ -1,54 +1,58 @@
-{-# LANGUAGE DeriveGeneric         #-}
-{-# LANGUAGE FlexibleContexts      #-}
-{-# LANGUAGE LambdaCase            #-}
-{-# LANGUAGE RankNTypes            #-}
-{-# LANGUAGE PartialTypeSignatures #-}
-{-# LANGUAGE ScopedTypeVariables   #-}
+{-# LANGUAGE CPP           #-}
+{-# LANGUAGE DeriveGeneric #-}
 
 {-# OPTIONS_GHC -Wno-unused-top-binds #-}
 {-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
 
 module Test.Control.Monad.IOSimPOR (tests) where
 
-import           Data.Fixed (Micro)
-import           Data.Foldable (foldl', traverse_)
-import           Data.Functor (($>))
-import qualified Data.List as List
-import           Data.Map (Map)
-import qualified Data.Map as Map
-import           Data.STRef.Lazy
+import Data.Fixed (Micro)
+#if __GLASGOW_HASKELL__ >= 910
+import Data.Foldable (traverse_)
+#else
+import Data.Foldable (foldl', traverse_)
+#endif
+import Data.Functor (($>))
+import Data.List qualified as List
+import Data.Map (Map)
+import Data.Map qualified as Map
 
-import           System.Exit
-import           System.IO.Error (ioeGetErrorString, isUserError)
+import System.Exit
+import System.IO.Error (ioeGetErrorString, isUserError)
 
-import           Control.Exception (ArithException (..), AsyncException)
-import           Control.Monad
-import           Control.Monad.Fix
-import           Control.Monad.ST.Lazy (ST, runST)
+import Control.Exception (ArithException (..), AsyncException)
+import Control.Monad
+import Control.Monad.Fix
 
-import           Control.Concurrent.Class.MonadSTM
-import           Control.Monad.Class.MonadAsync
-import           Control.Monad.Class.MonadFork
-import           Control.Monad.Class.MonadSay
-import           Control.Monad.Class.MonadTest
-import           Control.Monad.Class.MonadThrow
-import           Control.Monad.Class.MonadTime.SI
-import           Control.Monad.Class.MonadTimer.SI
-import           Control.Monad.IOSim
+import Control.Concurrent.Class.MonadSTM
+import Control.Monad.Class.MonadAsync
+import Control.Monad.Class.MonadFork
+import Control.Monad.Class.MonadSay
+import Control.Monad.Class.MonadTest
+import Control.Monad.Class.MonadThrow
+import Control.Monad.Class.MonadTime.SI
+import Control.Monad.Class.MonadTimer.SI
+import Control.Monad.IOSim
 
-import           GHC.Generics
+import GHC.Generics
 
-import           Test.Control.Monad.IOSim (TimeoutDuration, ActionDuration,
-                   WithSanityCheck (..), ignoreSanityCheck,
-                   isSanityCheckIgnored, singleTimeoutExperiment,
-                   withSanityCheck)
-import           Test.Control.Monad.Utils
-import           Test.Control.Monad.STM
+import Test.Control.Monad.IOSim (ActionDuration, TimeoutDuration,
+           WithSanityCheck (..), ignoreSanityCheck, isSanityCheckIgnored,
+           singleTimeoutExperiment, withSanityCheck)
+import Test.Control.Monad.STM
+import Test.Control.Monad.Utils
 
-import           Test.QuickCheck
-import           Test.Tasty (TestTree, testGroup)
-import           Test.Tasty.QuickCheck
-import qualified Data.List.Trace as Trace
+import Data.List.Trace qualified as Trace
+import Test.QuickCheck
+import Test.Tasty (TestTree, testGroup)
+import Test.Tasty.QuickCheck
+
+import System.IO.Unsafe
+
+import Data.IORef
+import Data.Time (secondsToDiffTime)
+import Data.Void (Void)
+import GHC.Conc (pseq)
 
 tests :: TestTree
 tests =
@@ -71,7 +75,9 @@ tests =
       , testProperty "timeout"                prop_timeout
       , testProperty "timeouts"               prop_timeouts
       , testProperty "stacked timeouts"       prop_stacked_timeouts
+      , testProperty "{register,thread}Delay" unit_registerDelay_threadDelay
       ]
+    , testProperty "infinite simulation"      prop_explore_endless_simulation
     , testProperty "threadId order (IOSim)"   (withMaxSuccess 1000 prop_threadId_order_order_Sim)
     , testProperty "forkIO order (IOSim)"     (withMaxSuccess 1000 prop_fork_order_ST)
     , testGroup "throw/catch unit tests"
@@ -339,23 +345,22 @@ propSimulates cmp (Shrink2 (Tasks tasks)) =
       Left (FailureInternal msg)
                   -> counterexample msg False
       Left x      -> counterexample (ppTrace trace)
-                   $ counterexample (show x) True 
+                   $ counterexample (show x) True
 
 -- NOTE: This property needs to be executed sequentially, otherwise it fails
 -- undeterministically, which `exploreSimTraceST` does.
 --
-propExploration :: Compare -> (Shrink2 Tasks) -> Property
+propExploration :: Compare -> Shrink2 Tasks -> Property
 propExploration cmp (Shrink2 ts@(Tasks tasks)) =
-  any (not . null . (\(Task steps)->steps)) tasks ==>
-    runST $
-    -- traceNoDuplicates $ \addTrace->
-    exploreSimTraceST (\a -> a { explorationDebugLevel = 0 })
-                      (say (show ts) >> runTasks cmp tasks)
-                    $ \_ trace -> do
-    -- addTrace trace
+  not (all (null . (\(Task steps)->steps)) tasks) ==>
+    traceNoDuplicates $ \addTrace->
+    exploreSimTrace (\a -> a { explorationDebugLevel = 0 })
+                    (say (show ts) >> runTasks cmp tasks)
+                    $ \_ trace ->
+    addTrace trace
     -- TODO: for now @coot is leaving `Trace.ppTrace`, but once we change all
     -- assertions into `FailureInternal`, we can use `ppTrace` instead
-    return $ counterexample (Trace.ppTrace show (ppSimEvent 0 0 0) trace) $
+    counterexample (Trace.ppTrace show (ppSimEvent 0 0 0) trace) $
              case traceResult False trace of
                Right (m,a) -> property (m >= a)
                Left (FailureInternal msg)
@@ -410,31 +415,81 @@ unit_issue113_nonDep =
 -- Test manually, and supply a small value of n.
 propPermutations :: Int -> Property
 propPermutations n =
-  runST $
   traceNoDuplicates $ \addTrace ->
-  exploreSimTraceST (withScheduleBound 10000) (doit n) $ \_ trace -> do
-    addTrace trace
+  exploreSimTrace (withScheduleBound 10000) (doit n) $ \_ trace ->
+    addTrace trace $
     let Right result = traceResult False trace
-    return $ tabulate "Result" [show $ result] $ True
+     in tabulate "Result" [show result] True
 
 doit :: Int -> IOSim s [Int]
 doit n = do
-  r <- atomically $ newTVar []
+  r <- newTVarIO []
   exploreRaces
   mapM_ (\i -> forkIO $ atomically $ modifyTVar r (++[i])) [1..n]
   threadDelay 1
-  atomically $ readTVar r
+  readTVarIO r
+
+traceNoDuplicates :: forall a b.
+                     (Show a)
+                  => ((a -> b -> b) -> Property)
+                  -> Property
+-- this NOINLINE pragma is useful for debugging if `r` didn't flow outside of
+-- `traceNoDuplicate`.
+{-# NOINLINE traceNoDuplicates #-}
+traceNoDuplicates k = unsafePerformIO $ do
+    r <- newIORef (Map.empty :: Map String Int)
+    return $ r `pseq`
+             (k (addTrace r) .&&. counterexample "trace counts" (maximum (Map.elems (traceCounts r)) === 1))
+  where
+    addTrace :: IORef (Map String Int) -> a -> b -> b
+    addTrace r t x = unsafePerformIO $ do
+      let s = show t
+      atomicModifyIORef r
+        (\m->
+          let m' = Map.insertWith (+) s 1 m
+          in (m', ())
+        )
+      return x
+
+    traceCounts :: IORef (Map String Int) -> Map String Int
+    traceCounts r = unsafePerformIO $ readIORef r
 
 
-traceNoDuplicates :: forall s a prop. (Show a, Testable prop)
-                  => ((a -> ST s ()) -> ST s prop)
-                  -> ST s Property
-traceNoDuplicates k = do
-  v <- newSTRef Map.empty :: ST s (STRef s (Map String Int))
-  prop <- k (\a -> modifySTRef v (Map.insertWith (+) (show a) 1))
-  m <- readSTRef v
-  return $ prop .&&. counterexample (show (Map.keys $ Map.filter (> 1) m)) (maximum m === 1)
+-- | Checks that IOSimPOR is capable of analysing an infinite simulation
+-- lazily.
+--
+-- If the test fails with "outer timeout" error it means that 'exploreSimTrace'
+-- is not lazy and there's something wrong with IOSimPOR implementation and
+-- the trace is being forced way too soon. If, on the other hand, this property
+-- fails with "inner timeout" error, it means that 'exploreSimTrace' is somewhat
+-- lazy, however the trace is not being lazily generated.
+--
+--
+prop_explore_endless_simulation :: Positive (Small Integer) -> Property
+prop_explore_endless_simulation (Positive (Small finalTime)) =
+    counterexample "outer timeout"
+  $ label (show (secondsToDiffTime finalTime))
+  $ within 15000000 -- 15 seconds
+  $ exploreSimTrace
+      id sim $ \_ trace -> do
+        let l = takeWhile (\(t, _, _, _) -> t < Time (secondsToDiffTime finalTime))
+              . traceEvents
+              $ trace
+         in ioProperty $ do
+           r <- timeout 10 $ evaluate (foldl' (flip seq) True l)
+           case r of
+             Nothing -> return $ counterexample "inner timeout" False
+             Just _  -> return (property True)
+  where
+    thread :: forall s. IOSim s Void
+    thread = do
+      threadDelay 1
+      thread
 
+    sim :: forall s. IOSim s Void
+    sim = do
+      exploreRaces
+      withAsync thread wait
 
 --
 -- IOSim reused properties
@@ -446,11 +501,10 @@ traceNoDuplicates k = do
 
 prop_stm_graph_sim :: TestThreadGraph -> Property
 prop_stm_graph_sim g =
-  runST $
   traceNoDuplicates $ \addTrace ->
-    exploreSimTraceST id (prop_stm_graph g) $ \_ trace -> do
+    exploreSimTrace id (prop_stm_graph g) $ \_ trace -> do
       addTrace trace
-      return $ counterexample (Trace.ppTrace show show trace) $
+      $ counterexample (Trace.ppTrace show show trace) $
         case traceResult False trace of
           Right () -> property True
           Left e   -> counterexample (show e) False
@@ -498,7 +552,7 @@ prop_mfix_purity (Positive n) =
        Left e  -> counterexample (show e) False
   where
     factorial :: (Int -> Int) -> Int -> Int
-    factorial = \rec_ k -> if k <= 1 then 1 else k * rec_ (k - 1)
+    factorial rec_ k = if k <= 1 then 1 else k * rec_ (k - 1)
 
 prop_mfix_purity_2 :: [Positive Int] -> Property
 prop_mfix_purity_2 as =
@@ -899,7 +953,7 @@ prop_timeout
     :: TimeoutDuration
     -> ActionDuration
     -> Property
-prop_timeout intendedTimeoutDuration intendedActionDuration = 
+prop_timeout intendedTimeoutDuration intendedActionDuration =
     exploreSimTrace id experiment $ \_ trace ->
         case traceResult False trace of
           Right a -> a
@@ -971,6 +1025,35 @@ prop_stacked_timeouts timeout0 timeout1 actionDuration =
               | otherwise -- i.e. timeout0 >= timeout1
               = Just Nothing
 
+unit_registerDelay_threadDelay :: Property
+unit_registerDelay_threadDelay =
+    exploreSimTrace id experiment $ \_ trace ->
+        counterexample (ppTrace_ trace)
+      . either (\e -> counterexample (show e) False) id
+      . traceResult False
+      $ trace
+  where
+    experiment :: IOSim s Property
+    experiment = do
+      v0 <- registerDelay 2
+      v1 <- newTVarIO False
+
+      _ <- forkIO $ do
+        threadDelay 1
+        atomically $ writeTVar v1 True
+
+      atomically $ do
+        b0 <- readTVar v0
+        b1 <- readTVar v1
+        check $ b0 || b1
+
+      let delay = 2
+      t0 <- getMonotonicTime
+      threadDelay delay
+      t1 <- getMonotonicTime
+
+      return (t1 `diffTime` t0 === delay)
+
 unit_timeouts_and_async_exceptions_1 :: Property
 unit_timeouts_and_async_exceptions_1 =
     exploreSimTrace id experiment $ \_ trace ->
@@ -990,7 +1073,7 @@ unit_timeouts_and_async_exceptions_1 =
       threadDelay (delay / 2)
       killThread tid
       threadDelay 1
-      return $ property True 
+      return $ property True
 
 
 unit_timeouts_and_async_exceptions_2 :: Property
@@ -1012,7 +1095,7 @@ unit_timeouts_and_async_exceptions_2 =
       threadDelay (delay / 2)
       killThread tid
       threadDelay 1
-      return $ property True 
+      return $ property True
 
 
 unit_timeouts_and_async_exceptions_3 :: Property
@@ -1034,7 +1117,7 @@ unit_timeouts_and_async_exceptions_3 =
       threadDelay (delay / 2)
       killThread tid
       threadDelay 1
-      return $ property True 
+      return $ property True
 
 --
 -- MonadMask properties
